import React, { useEffect, useState } from 'react'
import './App.css'

type Symbol = 'X' | 'O';

type CellState = {
  symbol?: Symbol;
  isFading?: boolean;
};

type RowState = CellState[];
type BoardState = RowState[];

type CellClickHandler = (row: number, col: number) => void;
type CellProps = CellState & { rowIndex: number, columnIndex: number, onClick?: CellClickHandler };

function Cell({ rowIndex, columnIndex, symbol, isFading, onClick }: CellProps) {
  return <div className={`board-cell ${isFading ? 'fading' : ''}`} onClick={() => onClick && onClick(rowIndex, columnIndex)}>{symbol}</div>;
}

type RowProps = {
  index: number,
  state: RowState
  onCellClick: CellClickHandler;
};

function Row({ state, index, onCellClick }: RowProps) {
  return (
    <div className="board-row">
      {state.map((cellState: CellState, i) => <Cell
        key={`cell-${index}-${i}`}
        rowIndex={index}
        columnIndex={i}
        {...cellState}
        onClick={onCellClick}/>)}
    </div>
  )
}

type BoardProps = {
  state: BoardState;
  hasWinner?: boolean;
  onCellClick: CellClickHandler;
}

function Board({ state, hasWinner, onCellClick }: BoardProps) {
  return (
    <div className={`board ${hasWinner ? 'game-set' : ''}`}>
      {state.map((row, i) => <Row key={`row-${i}`} index={i} state={row} onCellClick={onCellClick}/>)}
    </div>
  )
}

function hasWin(board: BoardState, size: number, row: number, column: number): boolean {
  const targetCell = board[row]?.[column];
  if (!targetCell) { return false; }

  let win = true;
  for (let r = 0; r < size; r++) {
    if (board[r]?.[column]?.symbol !== targetCell.symbol) {
      win = false;
      break;
    }
  }
  if (win) {
    return true;
  }

  win = true;
  for (let c = 0; c < size; c++) {
    if (board[row]?.[c]?.symbol !== targetCell.symbol) {
      win = false;
      break;
    }
  }
  if (win) {
    return true;
  }
  if (row !== column) {
    return false;
  }


  win = true;
  for (let i = 0; i < size; i++) {
    if (board[i]?.[i]?.symbol !== targetCell.symbol) {
      win = false;
      break;
    }
  }

  return win;
}

type MoveRecord = {
  row: number;
  col: number;
  player: Symbol;
};

type GameState = {
  board: BoardState;
  latestMoves:MoveRecord[];
  currentPlayer: Symbol;
  winner?: Symbol;
}

function useBoardState(size: number = 3): [GameState, { reset: ()=> void; playOn: (row: number, col: number) => boolean } ]  {
  const [state, updateState] = useState<GameState>({
    board: Array(size).fill(Array(size).fill({})),
    latestMoves: [],
    currentPlayer: 'O',
  });

  return [
    state,
    {
      reset: () => updateState({
        board: Array(size).fill(Array(size).fill({})),
        latestMoves: [],
        currentPlayer: 'O',
      }),

      playOn: (row: number, col: number) => {
        if (state.winner) {
          return false;
        }

        const targetCell = state.board[row]?.[col];
        if (!targetCell || targetCell.symbol) {
          return false;
        }

        const cellToFading = state.latestMoves.length >= (2*size-1) ? state.latestMoves[0] : undefined;
        const fadingRow = cellToFading?.row;
        const fadingCol = cellToFading?.col;

        const nextState: GameState = {
          board: state.board.map((oldRow, r) => oldRow.map((oldCell, c)=> {
            return {
              ...(oldCell.isFading ? {} : oldCell),
              ...(r===row && c===col ? { symbol: state.currentPlayer } : undefined),
              ...(r===fadingRow && c===fadingCol ? { isFading: true} : undefined),
            };
          })),

          latestMoves: [
            ...state.latestMoves,
            {
              row,
              col,
              player: state.currentPlayer,
            }
          ].slice(-2*size+1),

          currentPlayer: state.currentPlayer === 'X' ? 'O' : 'X' as Symbol,
        };

        if (hasWin(nextState.board, size, row, col)) {
          nextState.winner = state.currentPlayer;
        }
        updateState(nextState);

        return true;
      },
    },
  ];
}
function App() {
  const [state, { playOn, reset: resetGame }] = useBoardState(5);

  const nextPlayer = state.currentPlayer;
  const latestCells = state.latestMoves;
  const winner = state.winner;

  return (
    <>
      <Board state={state.board} hasWinner={!!winner } onCellClick={playOn}/>
      <div className="messages">
        <p>
          <span style={{marginRight: '1em'}}>{winner ? `Winner is ${winner}` : (<>Next player is <em>{nextPlayer}</em></>)}</span>
          {latestCells.length === 0 && ("Click any cell to start")}
          {winner && <button onClick={resetGame}>Click to restart</button>}
        </p>
      </div>
    </>
  )
}

export default App
